@@ -1,4 +1,19 @@
-/* dirent.h for vms */
+/* dirent.h for vms
+Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
+2007, 2008, 2009 Free Software Foundation, Inc.
+This file is part of GNU Make.
+
+GNU Make is free software; you can redistribute it and/or modify it under the
+terms of the GNU General Public License as published by the Free Software
+Foundation; either version 3 of the License, or (at your option) any later
+version.
+
+GNU Make is distributed in the hope that it will be useful, but WITHOUT ANY
+WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
+A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License along with
+this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #ifndef VMSDIR_H
 #define VMSDIR_H
@@ -52,10 +67,11 @@ typedef struct DIR
 #define NULL 0
 #endif
 
-extern	DIR *opendir PARAMS (());
-extern	struct direct *readdir PARAMS ((DIR *dfd));
 #define rewinddir(dirp)	seekdir((dirp), (long)0)
-extern	int closedir PARAMS ((DIR *dfd));
-extern char *vmsify PARAMS ((char *name, int type));
+
+DIR *opendir ();
+struct direct *readdir (DIR *dfd);
+int closedir (DIR *dfd);
+const char *vmsify (const char *name, int type);
 
 #endif /* VMSDIR_H */
